package name.panitz;
import java.util.function.LongBinaryOperator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

public class Interpreter{
  public static enum Op{plus,minus,div,mod,mult;
    public LongBinaryOperator fun(){
      return switch(this) {
        case plus  -> (x,y)->x+y;
        case minus -> (x,y)->x-y;
        case div   -> (x,y)->x/y;
        case mod   -> (x,y)->x%y;
        case mult  -> (x,y)->x*y;
      };
    }
  }
  public static record Var(String name){}
  public static record Num(long n){}
  public static record OpEx(Object l,Op op, Object r){}
  public static record Assign(String name,Object o){}
  public static record If(Object cond,Object then,Object otherwise){}
  public static record While(Object cond,Object body){}
  public static record Sequence(List<Object> sts){}

  Map<String,Long> env = new TreeMap<>();

  public long eval(Object o){
    if (o instanceof Num n)     return n.n;
    if (o instanceof Var v)     return env.get(v.name);
    if (o instanceof OpEx op)   
      return op.op.fun().applyAsLong(eval(op.l),eval(op.r));
    if (o instanceof Assign as){
      var r = eval(as.o);
      env.put(as.name,r);
      return r;
    }
    if (o instanceof If ife)
      return eval(ife.cond)!=0?eval(ife.then):eval(ife.otherwise);
    if (o instanceof While w) {
      var r=0L;
      while(eval(w.cond)!=0)
        r=eval(w.body);
        return r;
    }
    if (o instanceof Sequence xs)
      return xs.sts.stream().reduce(0L,(r,x) -> eval(x),(x,y) -> y); 	
    throw new RuntimeException("Unknown expression: "+o); 
  } 
}