

package name.panitz.util;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Consumer;
import java.util.Comparator;

import java.util.NoSuchElementException;

public sealed interface LL<E> permits LL.Nil, LL.Cons{

  record Nil<E>() implements LL<E>{
    @Override public String toString(){return "[]";}
  }
 
  record Cons<E>(E hd,LL<E> tl) implements LL<E>{
    @Override public String toString(){
      return show(true,new StringBuffer("["));
    } 
  }

  default boolean isEmpty(){return this instanceof Nil;}

  default E head(){
    if (this instanceof Cons<E> c) return c.hd();
    throw new NoSuchElementException("head on empty list");
  }

  default LL<E> tail(){
    if (this instanceof Cons<E> c) return c.tl();
    throw new NoSuchElementException("tail on empty list");
  }

  @SuppressWarnings("rawtypes")
  final LL nil = new Nil<>();
  @SuppressWarnings("unchecked")
  static <E>LL<E> nil(){return nil;}

  static <E>LL<E> cons(E hd,LL<E> tl){return new Cons<>(hd,tl);}
  
  @SafeVarargs
  static <E>LL<E> of(E...es){
    LL<E> r = nil();
    for (int i=es.length-1;i>=0;i--) r = cons(es[i],r);
    return r; 
  }

  default String show(boolean first,StringBuffer result){
    if (isEmpty()) {
      result.append("]");
      return result.toString();
    }
    if (!first)result.append(", ");
    result.append(head());
    return tail().show(false,result);
  }


  static int binSearch(int x,int[] xs){
    var low = 0;
    var up  = xs.length-1;
    while (low<=up){
      var middle = (low+up)/2;
      if (xs[middle]==x) return middle;
      if (low==middle) low=up;
      if (low==up) break;
      if (xs[middle]>x) up = middle;
      else if (xs[middle]<x) low = middle;
    }
    return -1;
  }
  

  default int length(){
    if (isEmpty())return 0;
    return 1+tail().length();
  }
  default int lengthIt(){
    var r = 0;
    for (var it=this;!it.isEmpty();it=it.tail())
      r++;
    return r;
  }

  @SuppressWarnings("preview")
  default E last(){
    return switch (this) {
    case Cons<E>(E x,Nil<E>()) -> x;
    case Cons<E> xs -> xs.tail().last();
    case Nil<E>() -> throw new NoSuchElementException();
    };  
  }

  default LL<E> append(LL<E> that){
    if (this instanceof Cons) return cons(head(),tail().append(that)); 
    else return that;
  }


  default LL<E> drop(int i){
    if (this instanceof Cons<E> c && i>0)
      return c.tail().drop(i-1);
    return this;
  }
   
  default LL<E> take(int i){
    if (this instanceof Cons && i>0)
      return cons(head(),tail().take(i-1));
    return nil();
  }

  default LL<E> sublist(int from, int length) {
    return drop(from).take(length);
  }

}
